//=================================================
interface Point2D  {
	Object clone();
	boolean equals(Object o);
	String toString();
	double getX();
	double getY();
	void setX(double x);
	void setY(double y);
	void translate(double dx, double dy);
}
//=================================================
abstract class Point2DAbstract  implements Point2D {	
	abstract public Object clone();	
	public boolean equals(Object o) {
	    System.out.println("equals Point2DAbstract sur "+ this);
		if ( o instanceof Point2D ) {
			Point2D p = (Point2D) o;
			return getX() == p.getX() && getY() == p.getY();
		}
		else
			return false;
	}
	public String toString() {
		return "(" + getX() + "," + getY() + ")";
	}
	public void translate(double dx, double dy) {
		setX(getX() + dx);
		setY(getY() + dy);
	}
}
//=================================================
/**
 *
 */
class Point2DArray extends Point2DAbstract {
	/**
	 * Constructeur appelé à l'instanciation des objets de type Point
	 */
	public Point2DArray(double x, double y) {
		coords = new double[2];
		coords[0] = x;
		coords[1] = y;
	}
	//Copie
	public Object clone() {
		return new Point2DArray(coords[0], coords[1]);
	}
	//Accesseurs
	public double getX() {
		return coords[0];
	}
	public double getY() {
		return coords[1];
	}
	public void setX(double x) {
		coords[0] = x;
	}
	public void setY(double y) {
		coords[1] = y;
	}
	private double coords [];
}
//=================================================
/**
 *
 */
class Point2DXY extends Point2DAbstract {
	/**
	 * Constructeur appelé à l'instanciation des objets de type Point
	 */
	public Point2DXY(double x, double y) {
		this.x = x;
		this.y = y;
	}
	//Copie
	public Object clone() {
		return new Point2DXY(x, y);
	}
	//Accesseurs
	public double getX() {
		return x;
	}
	public double getY() {
		return y;
	}
	public void setX(double x) {
		this.x = x;
	}
	public void setY(double y) {
		this.y = y;
	}
	private double x;
	private double y;
}
//==========================================================
interface Color {
	Object clone();
	boolean equals(Object o);
	String toString();
	int getR();
	int getG();
	int getB();
	int getA();
	void set(int r, int g, int b, int a);
	void setR(int r);
	void setG(int g);
	void setB(int b);
	void setA(int a);
}
//=================================================
abstract class ColorAbstract  implements Color {
	abstract public Object clone();
	public boolean equals(Object o) {
	    System.out.println("equals ColorAbstract sur "+ this);
		if ( o instanceof Color )
		{
			Color c = (Color) o;
			return getR() == c.getR() && getG() == c.getG() && getB() == c.getB() && getA() == c.getA();
		}
		else
			return false;
	}
	public String toString() {
		return "(" + getR() + "," + getG() + "," + getB() + "," + getA() + ")";
	}
	public void set(int r, int g, int b, int a) {
		setR(r);
		setG(g);
		setB(b);
		setA(a);
	}

}
//=================================================
class ColorInt extends ColorAbstract {
	public ColorInt(int r, int g, int b, int a){
		set(r, g, b, a);
	}
	public Object clone() {
		return new ColorInt(getR(), getG(), getB(), getA());
	}
	public int getR() {
		return (int)((_col & 0xFF000000) >> 24);
	}
	public int getG() {
		return (int)((_col & 0x00FF0000) >> 16);
	}
	public int getB() {
		return (int)((_col & 0x0000FF00) >> 8);
	}
	public int getA() {
		return (int)((_col & 0x000000FF));
	}
	public void setR(int r) {
		_col = _col & 0x00FFFFFF;
		_col = _col | ((int)r) << 24;
	}
	public void setG(int g) {
		_col = _col & 0xFF00FFFF;
		_col = _col | ((int)g) << 16;
	}
	public void setB(int b) {
		_col = _col & 0xFFFF00FF;
		_col = _col | ((int)b) << 8;
	}
	public void setA(int a) {
		_col = _col & 0xFFFFFF00;
		_col = _col | ((int)a);
	}
	private int _col;
}
//=================================================
class Color4Int extends ColorAbstract {
	public Color4Int (int r, int g, int b, int a){
		_r = r;
		_g = g;
		_b = b;
		_a = a;
	}
	public Object clone() {
		return new Color4Int(_r, _g, _b, _a);
	}
	public int getR() {
		return _r;
	}
	public int getG() {
		return _g;
	}
	public int getB() {
		return _b;
	}
	public int getA() {
		return _a;
	}
	public void setR(int r) {
		_r = r;
	}
	public void setG(int g) {
		_g = g;
	}
	public void setB(int b) {
		_b = b;
	}
	public void setA(int a) {
		_a = a;
	}
	private int _r, _g, _b, _a;
}




/*
 * Solution de delegation par re-implementation complete des methodes des interfaces
 */
//=================================================
class PointWithColor_1 implements Color, Point2D {
	public PointWithColor_1(double x, double y, int r, int g, int b, int a) {
		coords = new double[2];
		coords[0] = x;
		coords[1] = y;
		set(r,g,b,a);
	}
	public Object clone() {
		return new PointWithColor_1(coords[0], coords[1], getR(), getG(), getB(), getA());
	}
	public boolean equals(Object o) {
		if ( o instanceof PointWithColor_1 )
		{
			PointWithColor_1 pc = (PointWithColor_1) o;
			return getR() == pc.getR() && getG() == pc.getG() 
					&& getB() == pc.getB() && getA() == pc.getA()
					&& getX() == pc.getX() && getY() == pc.getY();
		}
		else
			return false;
	}
	public String toString() {
		return "point (" + getX() + "," + getY() + ") with color (" + getR() + "," + getG() + "," + getB() + "," + getA() + ")";
	}
	public int getR() {
		return (int)((_col & 0xFF000000) >> 24);
	}
	public int getG() {
		return (int)((_col & 0x00FF0000) >> 16);
	}
	public int getB() {
		return (int)((_col & 0x0000FF00) >> 8);
	}
	public int getA() {
		return (int)((_col & 0x000000FF));
	}
	public void setR(int b) {
		_col = _col & 0x00FFFFFF;
		_col = _col | ((int)b) << 24;
	}
	public void setG(int b) {
		_col = _col & 0xFF00FFFF;
		_col = _col | ((int)b) << 16;
	}
	public void setB(int b) {
		_col = _col & 0xFFFF00FF;
		_col = _col | ((int)b) << 8;
	}
	public void setA(int b) {
		_col = _col & 0xFFFFFF00;
		_col = _col | ((int)b);
	}
	public void set(int r, int g, int b, int a) {
		setR(r);
		setG(g);
		setB(b);
		setA(a);
	}
	public double getX() {
		return coords[0];
	}
	public double getY() {
		return coords[1];
	}
	public void setX(double x) {
		coords[0] = x;
	}
	public void setY(double y) {
		coords[1] = y;
	}
	public void translate(double dx, double dy) {
		setX(getX() + dx);
		setY(getY() + dy);
	}
	private double coords [];
	private int _col;
}

/*
 * Solution de delegation par 2 donnees membres de types figes
 */
//=================================================
class PointWithColor_2 implements Color, Point2D {
	public PointWithColor_2(Point2DArray p, ColorInt c) {
		this.p = (Point2DArray)p.clone();
		this.c = (ColorInt)c.clone();
	}
	public PointWithColor_2(double x, double y, int r, int g, int b, int a) {
		p = new Point2DArray(x, y);
		c = new ColorInt(r,g,b,a);
	}
	public Object clone() {
		return new PointWithColor_2(p, c);
	}
	public boolean equals(Object o) {
		if ( o instanceof PointWithColor_2 )
		{
			PointWithColor_2 pc = (PointWithColor_2) o;
			return p.equals(pc.getP()) && c.equals(pc.getC());
		}
		else
			return false;
	}
	public String toString() {
		return "point " + p + " with color " + c;
	}
	public int getR() {
		return c.getR();
	}
	public int getG() {
		return c.getG();
	}
	public int getB() {
		return c.getB();
	}
	public int getA() {
		return c.getA();
	}
	public void setR(int b) {
		c.setR(b);
	}
	public void setG(int b) {
		c.setG(b);
	}
	public void setB(int b) {
		c.setB(b);
	}
	public void setA(int b) {
		c.setA(b);
	}
	public void set(int r, int g, int b, int a) {
		c.set(r, g, b, a);
	}
	public double getX() {
		return p.getX();
	}
	public double getY() {
		return p.getY();
	}
	public void setX(double x) {
		p.setX(x);
	}
	public void setY(double y) {
		p.setY(y);
	}
	private Point2DArray getP(){
		return (Point2DArray)p.clone();
	}
	private ColorInt getC(){
		return (ColorInt)c.clone();
	}
	public void translate(double dx, double dy) {
		p.translate(dx, dy);
	}
	private Point2DArray p;
	private ColorInt c;
}

/*
 * Solution de delegation par 2 donnees membres de types abstracts
 */
//=================================================
class PointWithColor_3 implements Color, Point2D {
	public PointWithColor_3(Point2D p, Color c) {
		this.p = (Point2D) p.clone();
		this.c = (Color) c.clone();
	}
	public Object clone() {
		return new PointWithColor_3(p, c);
	}
	public boolean equals(Object o) {
	        boolean testColor = true;
	        boolean testPoint = true;
                boolean testType  = false;
		if ( o instanceof Color ) {
		    testColor = c.equals((Color)o);
		    testType  = true;
		}
		if ( o instanceof Point2D ) {
		    testPoint = p.equals((Point2D)o);
		    testType  = true;
		}
		return testType && testColor && testPoint;
	}
	public String toString() {
		return "point " + p + " with color " + c;
	}
	public int getR() {
		return c.getR();
	}
	public int getG() {
		return c.getG();
	}
	public int getB() {
		return c.getB();
	}
	public int getA() {
		return c.getA();
	}
	public void setR(int b) {
		c.setR(b);
	}
	public void setG(int b) {
		c.setG(b);
	}
	public void setB(int b) {
		c.setB(b);
	}
	public void setA(int b) {
		c.setA(b);
	}
	public void set(int r, int g, int b, int a) {
		c.set(r, g, b, a);
	}
	public double getX() {
		return p.getX();
	}
	public double getY() {
		return p.getY();
	}
	public void setX(double x) {
		p.setX(x);
	}
	public void setY(double y) {
		p.setY(y);
	}
	private Point2D getP(){
		return p;
	}
	private Color getC(){
		return c;
	}
	public void translate(double dx, double dy) {
		p.translate(dx, dy);
	}
	private Point2D p;
	private Color c;
}

/**
 *
 */
public class Deleg_3cases {

	public static void main(String argv[]) {
		Point2D p  = new Point2DXY(5,5);
		Point2D p2 = new Point2DArray(5,5);
		Color a = new ColorInt(30,40,50,200);
		Color b = new Color4Int(30,40,50,200);
		System.out.println(a + " <==> " + b);
		
		PointWithColor_1 pc1 = new PointWithColor_1(10,10,32,42,52,100);
		PointWithColor_2 pc2 = new PointWithColor_2((Point2DArray)p2,(ColorInt)a);
		PointWithColor_3 pc3 = new PointWithColor_3(p2,a);
		PointWithColor_3 pc4 = new PointWithColor_3(p,b);
		System.out.println(pc1);
		System.out.println(pc2);
		System.out.println(pc3);
		System.out.println(pc4);
	}
}